///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2015 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////


using System;
using System.Collections.Generic;
using System.Data;

using com.espertech.esper.compat;
using com.espertech.esper.compat.collections;
using com.espertech.esper.util;

namespace com.espertech.esper.client
{
    /// <summary>
    /// Container for database configuration information, such as
    /// options around getting a database connection and options to control the lifecycle
    /// of connections and set connection parameters.
    /// </summary>

    [Serializable]
    public class ConfigurationDBRef
    {
        private ConnectionFactoryDesc connectionFactoryDesc;
        private readonly ConnectionSettings connectionSettings;
        private ConnectionLifecycleEnum connectionLifecycleEnum;
        private ConfigurationDataCache dataCacheDesc;
        private MetadataOriginEnum metadataOrigin;
        private ColumnChangeCaseEnum columnChangeCase;
        private readonly IDictionary<Type, Type> dataTypeMapping;

        /// <summary>
        /// Gets or sets the auto-commit connection settings for new connections to this database.
        /// </summary>

        virtual public bool? ConnectionAutoCommit
        {
            get { return connectionSettings.AutoCommit ; }
            set { connectionSettings.AutoCommit = value; }
        }

        /// <summary>
        /// Gets or sets the transaction isolation level on new connections created for this database.
        /// </summary>

        virtual public IsolationLevel ConnectionTransactionIsolation
        {
            get { return connectionSettings.TransactionIsolation.GetValueOrDefault(); }
            set { connectionSettings.TransactionIsolation = value; }
        }

        /// <summary>
        /// Gets or sets the catalog name for new connections created for this database.
        /// </summary>

        virtual public String ConnectionCatalog
        {
            get { return connectionSettings.Catalog; }
            set { connectionSettings.Catalog = value; }
        }

        /// <summary>
        /// Gets or sets the LRU cache to a given size for the database.
        /// </summary>

        virtual public int LRUCache
        {
            set { dataCacheDesc = new ConfigurationLRUCache(value); }
        }

        /// <summary>
        /// Ctor.
        /// </summary>

        public ConfigurationDBRef()
        {
            connectionLifecycleEnum = ConnectionLifecycleEnum.RETAIN;
            connectionSettings = new ConnectionSettings();
            metadataOrigin = MetadataOriginEnum.DEFAULT;
            columnChangeCase = ColumnChangeCaseEnum.NONE;
            dataTypeMapping = new Dictionary<Type, Type>();
        }

        /// <summary>
        /// Sets the database provider connection.
        /// </summary>
        /// <param name="driverName">Name of the driver.</param>
        /// <param name="properties">The properties.</param>

        public void SetDatabaseDriver(String driverName, Properties properties)
        {
            connectionFactoryDesc = new DbDriverFactoryConnection(driverName, properties);
        }

        /// <summary>
        /// Sets the database driver.
        /// </summary>
        /// <param name="dbDriverFactoryConnection">The db driver factory connection.</param>
        public void SetDatabaseDriver( DbDriverFactoryConnection dbDriverFactoryConnection )
        {
            connectionFactoryDesc = dbDriverFactoryConnection;    
        }

        /// <summary>
        /// Sets the database driver.
        /// </summary>
        /// <param name="dbDriverFactoryConnection">The db driver factory connection.</param>
        /// <param name="properties">The properties.</param>
        public void SetDatabaseDriver(DbDriverFactoryConnection dbDriverFactoryConnection, Properties properties)
        {
            connectionFactoryDesc = new DbDriverFactoryConnection(
                dbDriverFactoryConnection.Driver.GetType(),
                properties);
        }

        /// <summary>
        /// Sets the database driver.
        /// </summary>
        /// <param name="dbSpecification">The db specification.</param>
        public void SetDatabaseDriver(DbDriverConfiguration dbSpecification)
        {
            connectionFactoryDesc = new DbDriverFactoryConnection(dbSpecification);
        }

        /// <summary> Returns the connection settings for this database.</summary>
        /// <returns> connection settings
        /// </returns>

        public virtual ConnectionSettings ConnectionSettings
        {
            get { return connectionSettings; }
        }

        /// <summary>
        /// Gets or sets the setting to control whether a new connection is obtained
        /// for each lookup, or connections are retained between lookups. Controls
        /// whether a new connection is obtained for each lookup, or connections
        /// are retained between lookups.
        /// </summary>

        public virtual ConnectionLifecycleEnum ConnectionLifecycle
        {
            get { return connectionLifecycleEnum; }
            set { connectionLifecycleEnum = value; }
        }

        /// <summary>
        /// Gets the descriptor controlling connection creation settings.
        /// </summary>

        public virtual ConnectionFactoryDesc ConnectionFactoryDesc
        {
            get { return connectionFactoryDesc; }
            set { connectionFactoryDesc = value; }
        }

        /// <summary>
        /// Configures an expiry-time cache of the given maximum age in seconds and purge interval in seconds.
        /// <para>
        /// Specifies the cache reference type to be weak references. Weak reference cache entries become
        /// eligible for garbage collection and are removed from cache when the garbage collection requires so.
        /// </para>
        /// </summary>
        /// <param name="maxAgeSeconds">is the maximum number of seconds before a query result is considered stale (also known as time-to-live)
        /// </param>
        /// <param name="purgeIntervalSeconds">is the interval at which the engine purges stale data from the cache
        /// </param>

        public virtual void SetExpiryTimeCache(double maxAgeSeconds, double purgeIntervalSeconds)
        {
            dataCacheDesc = new ConfigurationExpiryTimeCache(maxAgeSeconds, purgeIntervalSeconds, ConfigurationCacheReferenceTypeHelper.GetDefault());
        }

        /// <summary>
        /// Configures an expiry-time cache of the given maximum age in seconds and purge interval in seconds. Also allows
        /// setting the reference type indicating whether garbage collection may remove entries from cache.
        /// </summary>
        /// <param name="maxAgeSeconds">the maximum number of seconds before a query result is considered stale (also known as time-to-live)</param>
        /// <param name="purgeIntervalSeconds">the interval at which the engine purges stale data from the cache.</param>
        /// <param name="cacheReferenceType">specifies the reference type to use</param>
        public virtual void SetExpiryTimeCache(double maxAgeSeconds, double purgeIntervalSeconds, ConfigurationCacheReferenceType cacheReferenceType)
        {
            dataCacheDesc = new ConfigurationExpiryTimeCache(maxAgeSeconds, purgeIntervalSeconds, cacheReferenceType);
        }

        /// <summary>
        /// Gets a query result data cache descriptor.
        /// </summary>
        public virtual ConfigurationDataCache DataCacheDesc
        {
            get { return dataCacheDesc; }
        }

        /// <summary>
        /// Adds the SQL types binding.
        /// </summary>
        /// <param name="sqlType">Type of the SQL.</param>
        /// <param name="desiredType">The desired type.</param>
        public void AddSqlTypeBinding(Type sqlType, Type desiredType)
        {
            DatabaseTypeEnum typeEnum = DatabaseTypeEnum.GetEnum(desiredType.FullName);
            if (typeEnum == null)
            {
                String supported = DatabaseTypeEnum.Values.Render();
                throw new ConfigurationException("Unsupported type '" + desiredType.FullName + "' when expecting any of: " + supported);
            }
            this.dataTypeMapping[sqlType] = desiredType;
        }

        /// <summary>
        /// Returns the mapping of types that the engine must perform
        /// when receiving output columns of that sql types.
        /// </summary>
        public IDictionary<Type, Type> DataTypeMapping
        {
            get { return dataTypeMapping; }
        }

        /// <summary>
        /// Returns an enumeration indicating how the engine retrieves metadata about the columns
        /// that a given SQL query returns.
        /// <para/>
        /// The engine requires to retrieve result column names and types in order to build a resulting
        /// event type and perform expression type checking.
        /// </summary>
        /// <returns>indication how to retrieve metadata</returns>
        public MetadataOriginEnum MetadataRetrievalEnum
        {
            get { return metadataOrigin; }
        }

        /// <summary>
        /// Gets and sets an indicator that indicates how the engine should retrieve
        /// metadata about the columns that a given SQL query returns.
        /// <para>
        /// The engine requires to retrieve result column names and types in order to build a resulting
        /// event type and perform expression type checking.
        /// </para>
        /// </summary>
        public MetadataOriginEnum MetadataOrigin
        {
            get { return metadataOrigin; }
            set { metadataOrigin = value; }
        }

        /// <summary>
        /// Gets or sets an enum value determining how the engine changes case
        /// on output column names returned from statement or statement result
        /// set metadata.
        /// </summary>
        /// <returns>change case enums</returns>
        public ColumnChangeCaseEnum ColumnChangeCase
        {
            get { return columnChangeCase; }
            set { columnChangeCase = value; }
        }

        /// <summary>
        /// Indicates how the engine retrieves metadata about a statement's output columns.
        /// </summary>
        [Serializable]
        public enum MetadataOriginEnum
        {
            /// <summary>
            /// By default, get output column metadata from the prepared statement., unless
            /// an Oracle connection class is used in which case the behavior is SAMPLE.
            /// </summary>
            DEFAULT,

            /// <summary>
            /// Always get output column metadata from the prepared statement regardless of what driver
            /// or connection is used.
            /// </summary>
            METADATA,

            /// <summary>
            /// Obtain output column metadata by executing a sample query statement at statement
            /// compilation time. The sample statement
            /// returns the same columns as the statement executed during event processing.
            /// See the documentation for the generation or specication of the sample query statement.
            /// </summary>
            SAMPLE
        }

        /// <summary>
        /// Controls how output column names get reflected in the event type.
        /// </summary>
        [Serializable]
        public enum ColumnChangeCaseEnum
        {
            /// <summary>
            /// Leave the column names the way the database driver represents the column.
            /// </summary>
            NONE,

            /// <summary>
            /// Change case to lowercase on any column names returned by statement metadata.
            /// </summary>
            LOWERCASE,

            /// <summary>
            /// Change case to uppercase on any column names returned by statement metadata.
            /// </summary>
            UPPERCASE
        }
    }
}
